#include "include.h"

#define SCALE_OPENGL 0.5

void setupGLMatrix(const NxVec3 &, const NxMat33 &);
void drawSphere(NxShape *);
void drawCube(NxShape *);
void drawShape(NxShape *);
void drawActor(NxActor *);
void renderActors();
void getPhysicsResults();
void startPhysics();
NxActor *createPlane(NxVec3, float);
NxActor *createCube(float, float, float);
NxActor *createSphere(float, float, float);
NxActor *createFluid(float, float, float);

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

extern GLuint	myProgram;

void initNx(void)
{
	gPhysicsSDK = NxCreatePhysicsSDK(NX_PHYSICS_SDK_VERSION);

	gPhysicsSDK->setParameter(NX_SKIN_WIDTH, 0.0);
	gPhysicsSDK->setParameter(NX_DEFAULT_SLEEP_LIN_VEL_SQUARED, 1.0);
	gPhysicsSDK->setParameter(NX_DEFAULT_SLEEP_ANG_VEL_SQUARED, 1.0);

	sceneDesc.gravity.set(0, -9.8f, 0);
	sceneDesc.collisionDetection = true;
	gScene = gPhysicsSDK->createScene(sceneDesc);

	NxMaterial		*defaultMaterial = gScene->getMaterialFromIndex(0); 
	defaultMaterial->setRestitution(0.5);
	defaultMaterial->setStaticFriction(2.0);
	defaultMaterial->setDynamicFriction(1.5);

	createPlane(NxVec3(0.0, 1.0, 0.0), -1.0);

	for (float z = 0.0; z < 30.0; z += 1.5)
		for (float y = 0.0; y < 750.0; y += 75.0)
			for (float x = 0.0; x < 30.0; x += 1.5)
				if ((int)(y/75.0)%2)
					createSphere(x, y, z);
				else
					createCube(x, y, z);
	if (gScene)
		startPhysics();
}

void setupGLMatrix(const NxVec3 &pos, const NxMat33 &orient)
{
	float glmat[16]; //4x4 column major matrix for OpenGL.
	orient.getColumnMajorStride4(&(glmat[0]));
	pos.get(&(glmat[12]));

	//clear the elements we don't need:
	glmat[3] = glmat[7] = glmat[11] = 0.0f;
	glmat[15] = 1.0f;

	glMultMatrixf(&(glmat[0]));
}

void drawSphere(NxShape *sphere)
{
	NxMat34 pose = sphere->getGlobalPose();

	glPushMatrix();
	setupGLMatrix(pose.t, pose.M);
	NxReal r = sphere->isSphere()->getRadius();
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texObject[0]);
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, texObject[1]);
	{
		int ret;
		ret = glGetUniformLocation(myProgram, "texture1"); 
		glUniform1i(ret, 1);
	}
	glCallList(dispListSphere);
	glPopMatrix();
}

void drawCube(NxShape *cube)
{
	NxMat34 pose = cube->getGlobalPose();

	glPushMatrix();
	setupGLMatrix(pose.t, pose.M);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texObject[3]);
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, texObject[4]);
	{
		int ret;
		ret = glGetUniformLocation(myProgram, "texture1"); 
		glUniform1i(ret, 1);
	}

	glCallList(dispListCube);
	glPopMatrix();
}

void drawShape(NxShape* shape)
{
	switch(shape->getType())
	{
	case NX_SHAPE_SPHERE:
		drawSphere(shape);
		break;
	case NX_SHAPE_BOX:
		drawCube(shape);
		break;
	}
}

void drawActor(NxActor* actor)
{
	NxShape*const* shapes = actor->getShapes();
	NxU32 nShapes = actor->getNbShapes();
	nShapes = actor->getNbShapes();
	while (nShapes--)
		drawShape(shapes[nShapes]);
}

void renderActors()
{
	int nbActors = gScene->getNbActors();
	NxActor** actors = gScene->getActors();
	while (nbActors--)
	{
		NxActor* actor = *actors++;
		drawActor(actor);
	}
}

void getPhysicsResults()
{
	// Get results from gScene->simulate(deltaTime)
	while (!gScene->fetchResults(NX_RIGID_BODY_FINISHED, false));
}

void startPhysics()
{
	NxReal deltaTime = 1.0f/60.0f;

	gScene->simulate(deltaTime);
	gScene->flushStream();
}

NxActor *createCube(float x, float y, float z)
{
	NxActorDesc		actorDesc;
	NxBodyDesc		bodyDesc;
	NxBoxShapeDesc		boxDesc;

	boxDesc.skinWidth = 0.0;
	boxDesc.dimensions.set(1.0 * SCALE_OPENGL, 1.0 * SCALE_OPENGL, 1.0 * SCALE_OPENGL);
	actorDesc.shapes.pushBack(&boxDesc);
	actorDesc.globalPose.t = NxVec3(x,y,z);
	actorDesc.body = &bodyDesc;
	actorDesc.density = 100;
	return gScene->createActor(actorDesc);
}

NxActor *createSphere(float x, float y, float z)
{
// Add a single-shape actor to the scene
	NxActorDesc	actorDesc;
	NxBodyDesc	bodyDesc;

	bodyDesc.angularVelocity = NxVec3(1.0f, 2.0f, 3.0f);
	NxSphereShapeDesc sphereDesc;
	sphereDesc.skinWidth = 0.0;
	sphereDesc.radius = 1.0 * SCALE_OPENGL;
	actorDesc.shapes.pushBack(&sphereDesc);
	
	actorDesc.globalPose.t = NxVec3(x,y,z);
	actorDesc.body = &bodyDesc;
	actorDesc.density = 100.0f;
	return gScene->createActor(actorDesc);
}

NxActor *createPlane(NxVec3 norm, float d)
{
	NxPlaneShapeDesc planeDesc;
	NxActorDesc actorDesc;

	planeDesc.skinWidth = 0.0;
	planeDesc.normal = norm;
	planeDesc.d = d;
	actorDesc.shapes.pushBack(&planeDesc);
	return gScene->createActor(actorDesc);
}

NxActor *createFluid(float x, float y, float z)
{
	NxActorDesc		actorDesc;
	NxBodyDesc		bodyDesc;
	NxFluid			*gFluid = NULL;
	NxFluidDesc		fluidDesc;
	NxFluidEmitter		*gFluidEmitter = NULL;
	NxFluidEmitterDesc	emitterDesc;
	NxVec3			*gParticles;
	NxReal			mat[] = {1,0,0,0, 0,0,1,0, 0,-1,0,0, 0,2,0,1};

	fluidDesc.setToDefault();
	fluidDesc.simulationMethod = NX_F_SPH;
	fluidDesc.staticCollisionRestitution = 0.1;
	fluidDesc.staticCollisionAdhesion = 0.5;
	fluidDesc.maxParticles = 100;
	fluidDesc.restParticlesPerMeter = 50;
	fluidDesc.stiffness = 1;
	fluidDesc.viscosity = 6;
	gParticles = new NxVec3[fluidDesc.maxParticles];
	fluidDesc.particlesWriteData.bufferPos = &gParticles[0].x;
	fluidDesc.particlesWriteData.bufferPosByteStride = sizeof(NxVec3);
	fluidDesc.particlesWriteData.maxParticles = fluidDesc.maxParticles;
	fluidDesc.particlesWriteData.numParticlesPtr = NULL; 
	gFluid = gScene->createFluid(fluidDesc);

	emitterDesc.flags |= NX_FEF_ENABLED;
	emitterDesc.setToDefault();
	emitterDesc.dimensionX = 0.03;
	emitterDesc.dimensionY = 0.03;
	emitterDesc.relPose.id();
	emitterDesc.relPose.setColumnMajor44(mat);
	emitterDesc.rate = 5.0;
	emitterDesc.randomAngle = 0.1f;
	emitterDesc.fluidVelocityMagnitude = 6.5f;
	emitterDesc.maxParticles = 100;
	emitterDesc.particleLifetime = 4.0f;
	emitterDesc.type = NX_FE_CONSTANT_FLOW_RATE;
	emitterDesc.shape = NX_FE_ELLIPSE;
	gFluidEmitter = gFluid->createEmitter(emitterDesc);

	bodyDesc.setToDefault();
	actorDesc.globalPose.t = NxVec3(x,y,z);
	actorDesc.body = &bodyDesc;
	return gScene->createActor(actorDesc);
}